package dataspork.reader;
import dataspork.tree.*;
import dataspork.reader.*;
import dataspork.interp.*;
import dataspork.util.SciStreamTokenizer;
import java.io.*;
import java.net.URL;

/* ClampsDataReader reads .sca files output by Clamps.  e.g.:

#RESTART
#PERIOD       10   steps
#NAMES time     accept   mc.pe.   
#UNITS secs.    ratio    null     
  0.2336091E-01  0.1666667E-01 -0.9036846E+01
  0.2380573E-01  0.2777778E-01 -0.8886460E+01
  0.2458148E-01  0.1944444E-01 -0.8729176E+01
  0.2433904E-01  0.3055556E-01 -0.8660258E+01
  0.2440499E-01  0.3333333E-01 -0.8412807E+01
  0.2497542E-01  0.3240741E-01 -0.8301323E+01

*/

public class ClampsDataReader implements DataReader {

  protected int ncols;
  final static public int MAX_COLS = 50;

  public ClampsDataReader() {}

  public DataTreeNode read(URL url) {

    ncols = 0;
    String filename = url.getFile();
    int k = filename.lastIndexOf('/');
    String shortFilename = filename.substring(++k);
    DataTreeNode parentnode = null;
    int nlabel = 0;
    int labelline = 0;
    int	lastTokenWasComment = 0;
    String[] colLabel = new String[MAX_COLS];
    StreamTokenizer st;

    String nameComment = "NAMES";

    // figure out how many columns data file has
    try {
      Reader r = new BufferedReader(new FileReader(filename));
      st = new SciStreamTokenizer(r);          
      st.eolIsSignificant(true);
      int token;

      // check for column labels (indicated by comment character)
      while ( (token=st.nextToken())!=st.TT_EOF && st.lineno() < 11) {
        if (st.ttype==st.TT_EOL) { continue; }
        if (st.lineno()==labelline && nlabel > 0) {
          colLabel[nlabel] = st.sval;
          nlabel++;
        }       

	if (lastTokenWasComment == 1) {
          if (st.sval.equals(nameComment)) {
	    labelline = st.lineno();
	    nlabel = 1;
	  }
	}

        if (token == '#') { 
	  lastTokenWasComment = 1;
	}
	else {
	  lastTokenWasComment = 0;
	}

        if (st.lineno() == 10) {
          ncols++;
        }
      }

      DataSource data;
      DataTreeNode node;
      InterpreterClass interp;
      parentnode = new DataTreeNode(shortFilename);
      String label;

      for (int i=1; i<=ncols; i++) {
        data = new AsciiColumnSource(filename,i);
        if (nlabel > 0 && colLabel[i] != null) { label = colLabel[i]; }
        else { label = "col " + i; }

        if (!label.equals("skip") && !label.equals("SKIP")) {

          node = new DataTreeNode(label,data);  
          node.setLongName(label+" ("+shortFilename+")");

          interp = new InterpreterClass(Class.forName("dataspork.interp.ScalarDataset"),node);
          node.addInterpreter(interp);
          if (node != null) parentnode.add(node);          
        }

      }
    } catch (Exception exc) {
      System.out.println("exception in AsciiDataReader: " + exc);
    }

    return parentnode;


  }
}
