/****************************************************************************
 * NCSA HDF                                                                 *
 * National Comptational Science Alliance                                   *
 * University of Illinois at Urbana-Champaign                               *
 * 605 E. Springfield, Champaign IL 61820                                   *
 *                                                                          *
 * For conditions of distribution and use, see the accompanying             *
 * hdf/COPYING file.                                                        *
 *                                                                          *
 ****************************************************************************/

package ncsa.hdf.message;

import java.awt.Dimension;
import ncsa.hdf.hdflib.HDFException;
import ncsa.hdf.hdflib.HDFLibrary;
import ncsa.hdf.hdflib.HDFConstants;
import ncsa.hdf.hdflib.HDFNativeData;

/**
 *  HDFSDS holds HDF SDS information

 *  @version   1.1.3 September 22 1997
 *  @author    Peter X. Cao
 */
public class HDFSDS extends HDFObject
{
    // The maximum value of the rank parameter
    public final static int MAX_VAR_DIMS = 32;

    /** the original image size */
    protected  Dimension originalImageSize;

    /** the requested image size */
    protected  Dimension selectedImageSize;

    /** dimension units */
    protected String[] dimUnits;

    /** dimension scales */
    protected Object[] dimScales;

    /** dimension types */
    protected int[] dimTypes;

    /** dimension names */
    protected String[] dimNames;

    /** dimension size */
    protected int[] dimSizes;

    /**
     *   hold image arguments.
     *  <pre>
     *   imageArgv[0] = number of images
     *   imageArgv[1] = image data type
     *  </pre>
     */
    protected  int imageArgv[];

    /** byte array containing data for animation images */
    protected  byte images[][];

    /**
     *   hold data range in the image
     *  <pre>
     *   dataRange[0] = the minimum the the image data
     *   dataRange[1] = the maxnimum the the image data
     *  </pre>
     */
    protected  double dataRange[];

    /**
     *  Constructs an HDFSDS with default parameters
     */
    public HDFSDS () { this(null, ""); }

    /**
     *  Construct an HDFSDS with specified HDFObjectNode and filename
     *
     *  @param node     the HDFObjectNode to be transfered to the server
     *  @param filename the requested HDF file
     */
    public HDFSDS (HDFObjectNode node, String filename)
    {
        super(node, filename);
        originalImageSize = null;
        selectedImageSize = null;
        data = null;
        imageArgv = null;
        dataRange = null;
        images = null;
        dimUnits = null;
        dimScales = null;
        dimTypes = null;
        dimNames = null;
        dimSizes = null;
    }

    /**
     *  create a message for sending to the client
     *
     *  @return  the HDFMessage created by this object
     */
    public HDFMessage toClient()
    {
        HDFMessage message  = new HDFMessage(getClass().getName());

        message.put("information", information);
        if (originalImageSize != null) message.put("originalImageSize", originalImageSize);
        if (selectedImageSize != null) message.put("selectedImageSize", selectedImageSize);
        if (data != null) message.put("data", data);
        if (imageArgv != null ) message.put("imageArgv", imageArgv);
        if (dataRange != null ) message.put("dataRange", dataRange);
        if (images != null ) message.put("images", images);
        if (dimUnits != null ) message.put("dimUnits", dimUnits);
        if (dimScales != null ) message.put("dimScales", dimScales);
        if (dimTypes != null ) message.put("dimTypes", dimTypes);
        if (dimNames != null) message.put("dimNames", dimNames);
        if (dimSizes != null) message.put("dimSizes", dimSizes);
        return message;
    }

    /**
     *  overwrite HDFGR
     */
    public void service()
    {
        //System.out.println("HDFSDS.service()");

        this.hdf = new HDFLibrary();
        try { information = getSDS (filename, nodeObject); }
        catch (Exception e) { information = "ERROR: exception in HDFSDS.getSDS()"; }
    }

    /**
     *  process the message received from the server
     *
     *  @param message  the HDFMessage received from the server
     */
    public void fromServer(HDFMessage message)
    {
        information = (String) message.get("information");
        originalImageSize = (Dimension) message.get("originalImageSize");
        selectedImageSize = (Dimension) message.get("selectedImageSize");
        data = message.get("data");
        imageArgv = (int[]) message.get("imageArgv");
        dataRange = (double[]) message.get("dataRange");
        images = (byte[][]) message.get("images");
        dimUnits = (String[]) message.get("dimUnits");
        dimScales = (Object[]) message.get("dimScales");
        dimTypes = (int[]) message.get("dimTypes");
        dimNames = (String[]) message.get("dimNames");
        dimSizes = (int[]) message.get("dimSizes");
    }
    /**
     *  get the selected HDF Generic raster image for the selected HDF file
     *
     *  @param filename   the string of the hdf file name
     *  @param node       the HDFObjectNode
     *  @return           the error message / SDS attributes
     *
     *  @exception ncsa.hdf.hdflib.HDFException
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     */
    private String getSDS (String filename, HDFObjectNode node)
        throws HDFException
    {
        //System.out.println("HDFSDS.getSDS()");

        int[] fileInfo = new int[2];
        String info = new String();
        int sdid = hdf.SDstart(filename, HDFConstants.DFACC_RDONLY);

        if (sdid == HDFConstants.FAIL)
        {
            hdf.SDend(sdid);
            return ("ERROR: native call to HDFLibrary.SDstart() failed.");
        }

        // get HDF SDS information
        if ( !hdf.SDfileinfo(sdid, fileInfo ) )
        {
            hdf.SDend(sdid);
            return ("ERROR: native call to HDFLibrary.SDfileinfo() failed.");
        }

        if (node.type == HDFObjectNode.SDSGLOBALATTR)
            info = getHdfSDSgattr(sdid, fileInfo);
        else if (node.type == HDFObjectNode.SDSDATASETATTR)
            info = getHdfSDSattr(sdid, fileInfo, node.index);
        else
            info = getSDSData(sdid, node.ref, node.plane);

        hdf.SDend(sdid);
        return info;
    }

    /**
     *  get the HDF file SDS image
     *
     *  @param sdid       the SDS identifier
     *  @param ref        the reference number of the selected HDF object node
     *  @return           the error message
     *
     *  @exception ncsa.hdf.hdflib.HDFException
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     */
    private String getSDSData(int sdid, int ref, int plane)
        throws HDFException
    {
        //System.out.println("HDFSDS.ggetSDSData()");

        int sdsid = -1;
        int index = -1;
        String retStr = "";

        index = hdf.SDreftoindex(sdid, ref);
        if (index == HDFConstants.FAIL)
            return "ERROR: native call to HDFLibrary.SDreftoindex() failed";

        if ((sdsid = hdf.SDselect(sdid,index)) == HDFConstants.FAIL)
        {
            hdf.SDendaccess(sdsid);
            return "ERROR: native call to HDFLibrary.SDselect() failed";
        }

        // get sds info.
        int [] SDInfo = new int[3];
        String datasetname[] = {""};
        int dimsize[] = new int[20]; // assume rank <= 20

        if (!hdf.SDgetinfo(sdsid, datasetname, dimsize, SDInfo))
        {
            hdf.SDendaccess(sdsid);
            return "ERROR: native call to HDFLibrary.SDgetinfo() failed";
        }

        //  type and rank of the data set
        int nt = SDInfo[1];
        int rank = SDInfo[0];

        if (rank <=0)
        {
            hdf.SDendaccess(sdsid);
            return "ERROR: rank from HDFLibrary.SDgetinfo() <= 0";
        }

        // get data again for the case of rank > 20
        if (rank > 20)
        {
            dimsize = new int[rank];
            hdf.SDgetinfo(sdsid, datasetname, dimsize, SDInfo);
        }

        // get the dimension information
        // load dimension information only once
        if (nodeObject.isPreview && plane < 2)
        {
            boolean ok = false;
            int dimid = -1;
            int[] dimInfo = new int[3];
            int max_length = 20; // Maximum length of label, units and format
            String dimName[] = {""};
            String[] dimStr = {"", "", ""}; //s[0] = label, s[1] = units[2] = format
            dimTypes = new int[rank];
            dimSizes = new int[rank];
            dimUnits = new String[rank];
            dimNames = new String[rank];
            dimScales = new Object[rank];

            for (int i=0; i<rank; i++)
            {
                dimid = hdf.SDgetdimid(sdsid, i);
                hdf.SDgetdimstrs(dimid, dimStr, max_length);
                hdf.SDdiminfo( dimid, dimName, dimInfo);

                dimUnits[i] = dimStr[1];
                dimNames[i] = dimName[0];
                dimSizes[i] = dimInfo[0];

                if (dimInfo[1]>0)
                {
                    dimTypes[i] = dimInfo[1];
                    dimScales[i] = defineDataObject(dimInfo[1], dimInfo[0]);
                    hdf.SDgetdimscale(dimid, dimScales[i]);
                }
                else
                {
                    dimTypes[i] = nt;
                    dimScales[i] = null;
                }
            }
        } //if (nodeObject.isPreview && plane < 2)

        if (nodeObject.selectedDimension == null)
        {
            nodeObject.selectedDimension = new int[3];
            nodeObject.selectedDimension[0] = 2;
            nodeObject.selectedDimension[1] = 1;
            nodeObject.selectedDimension[2] = 0;
        }

        if (nodeObject.depthRange == null)
        {
            nodeObject.depthRange = new int[2];
            nodeObject.depthRange[0] = 0;
            nodeObject.depthRange[1] = dimsize[nodeObject.selectedDimension[2]]-1;
        }

        // load SDS data
        int w=1, h=1;

        // set width and height
        if (rank == 1)
            w = dimsize[0];
        else if (rank == 2)
        {
            w = dimsize[1];
            h = dimsize[0];
        }
        else if (rank > 2)
        {
            w = dimsize[nodeObject.selectedDimension[0]];
            h = dimsize[nodeObject.selectedDimension[1]];
        }

        if ( (w < 0) || (h < 0) )
        {
            hdf.SDendaccess(sdsid);
            return ("ERROR: width/height is less than zero");
        }

        Dimension originalImageSize = new Dimension(w, h);

        // get the data range
        double maxmin[] = new double[2];
        double dataRange[] = null;
        if ( hdf.SDgetrange(sdsid, maxmin) )
        {
            dataRange = new double[2];
            dataRange[0] = maxmin[1];
            dataRange[1] = maxmin[0];
        }

        // set sds read options
        int start[] = new int[rank];
        int stride[]= new int[rank];
        int count[] = new int[rank];
        for (int i=0; i<rank; i++)
        {
            start[i] = 0;
            stride[i] = 1;
            count[i] = 1;
        }

        int orientation = nodeObject.selectedDimension[2];

        if ( nodeObject.isPreview )
       {
            // load preview data
            int skip_w = w / nodeObject.preferedImageSize + 1;
            int skip_h = h / nodeObject.preferedImageSize + 1;

            // update the actual size of loaded data
            w = Math.max(1, w/skip_w);
            h = Math.max(1, h/skip_h);

            // one dimension
            if (rank == 1)
            {
                start[0] = 0;
                stride[0] = skip_w;
                count[0] = Math.max(1, w);
            }
            else if (rank == 2)
            {
                start[0] = 0;
                start[1] = 0;
                stride[0] = skip_h;
                stride[1] = skip_w;
                count[0] = Math.max(1, h);
                count[1] = Math.max(1, w);
            }
            else if (rank > 2)
            {
                int x = nodeObject.selectedDimension[0];
                int y = nodeObject.selectedDimension[1];
                int z = nodeObject.selectedDimension[2];

                start[x]  = 0;
                stride[x] = skip_h;
                count[x]  = Math.max(1, dimsize[x]/skip_h);
                start[y]  = 0;
                stride[y] = skip_w;
                count[y]  = Math.max(1, dimsize[y]/skip_w);
                start[z]  = nodeObject.plane;
                stride[z] = 1;
                count[z]  = 1;
            }
        }
        else
        {
            // load full resolution data
            w = Math.max(1, nodeObject.selectedRange.width);
            h = Math.max(1, nodeObject.selectedRange.height);

            if (rank == 1)
            {
                // one dimension array
                start[0] = nodeObject.selectedRange.x;
                stride[0] = 1;
                count[0] = w;
            }
            else if (rank == 2)
            {
                // two dimension array
                start[0] = nodeObject.selectedRange.y;
                start[1] = nodeObject.selectedRange.x;
                stride[0] = 1;
                stride[1] = 1;
                count[0] = h  ;  // y
                count[1] = w  ;  //x
            }
            else if (rank >2)
            {
                // 3 or more dimension array
                start[nodeObject.selectedDimension[2]] = plane-1;
                start[nodeObject.selectedDimension[1]] = nodeObject.selectedRange.y;
                start[nodeObject.selectedDimension[0]] = nodeObject.selectedRange.x;
                count[nodeObject.selectedDimension[1]] = h; // y
                count[nodeObject.selectedDimension[0]] = w; // x
            }
        }  // end of load full resolution data setup

        int dataSize = w*h;
        Object data;
        int startPosition = nodeObject.depthRange[0];
        int endPosition = nodeObject.depthRange[1];
        int numberOfImages = endPosition-startPosition+1;

        // load all images for aninmation
        if ( (rank>2) && (plane > dimsize[orientation]) )
        {
            if (numberOfImages > 1)
            {
                this.images = new byte[numberOfImages-1][dataSize];

                /* The following code does not get the right order of data if buffSize > 1
                //load images buffSize at a time, this increases the memory
                // space but decreases the accessing time
                long freeMemory = Runtime.getRuntime().freeMemory();
                long needMemory = dataSize*HDFNativeData.getTypeSize(nt);
                int buffSize = (int)(freeMemory/needMemory*0.8); // leave 20% space for other use
                if (buffSize >= numberOfImages) buffSize = numberOfImages-1;
                if (buffSize < 1) buffSize = 1;
                data = defineDataObject(nt, dataSize*buffSize);
                int loop = (numberOfImages-1)/buffSize+1;
                int dataLeft = numberOfImages;
                int position = 0;
                for (int i=0; i<loop; i++)
                {
                    position = i*buffSize;
                    start[orientation] = startPosition+1+position;
                    if (dataLeft <= buffSize)
                    {
                        buffSize = dataLeft-1;
                        data = defineDataObject(nt, dataSize*buffSize);
                    }
                    count[orientation] = buffSize;
                    if ( !hdf.SDreaddata(sdsid,start,null,count,data) )
                    {
                        hdf.SDendaccess(sdsid);
                        return ("ERROR: native call to HDFLibrary.SDreaddata() failed.");
                    }
                    for (int j=0; j<buffSize; j++)
                    {
                       object2byte(data, nt, dataRange, j*dataSize, this.images[j+position]);
                    }
                    dataLeft = numberOfImages-(i+1)*buffSize;
                }
                data = null; //clean the memory
                */

                // load images one by one
                data = defineDataObject(nt, dataSize);
                for (int i=0; i<numberOfImages-1; i++)
                {
                    start[orientation] = startPosition+1+i;
                    if ( !hdf.SDreaddata(sdsid,start,null,count,data) )
                    {
                        hdf.SDendaccess(sdsid);
                        return ("ERROR: native call to HDFLibrary.SDreaddata() failed.");
                    }
                    object2byte(data, nt, dataRange, this.images[i]);
                }

            } // end of if (numberOfImages > 1)
        } // end of if ( (rank>2) && (plane > dimsize[orientation]) )
        else
        {
            // read data for the chosen plane
            data = defineDataObject(nt, dataSize);
            if ( !hdf.SDreaddata(sdsid,start,null,count,data) )
            {
                hdf.SDendaccess(sdsid);
                return ("ERROR: native call to HDFLibrary.SDreaddata() failed.");
            }

            this.originalImageSize = originalImageSize;
            this.selectedImageSize = new Dimension(w, h);
            this.data = data;
            int temp[] = {1, nt, rank};
            this.imageArgv = temp;
            if ( rank > 2 ) this.imageArgv[0] = numberOfImages;
            this.dataRange = dataRange;
        }

        hdf.SDendaccess(sdsid);
        return "";
    }


    /**
     * display the HDF file SDS global attributes
     *
     * @param sdid       the SDS identifier
     * @param fileInfo   the SDS data information
     * @return           the string of the global attributes.
     *
     *  @exception ncsa.hdf.hdflib.HDFException 
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     */
    private String getHdfSDSgattr ( int sdid, int[] fileInfo)
        throws HDFException
    {
        String textInfo = "";
        int sdsGlobalAttrNumber = fileInfo[1];

        textInfo = "\nGlobal Attributes";
        textInfo += "\n\nName";
        for (int j=0; j<MAX_NAME_LENGTH-4; j++) textInfo += " ";
        textInfo += " Value\n";
        for (int j=0; j<80; j++) textInfo += "_";

        // for each global attribute
        for (int i=0; i<sdsGlobalAttrNumber; i++)
        {
            // get attribute info.
            textInfo += "\n";
            String attrname = "";
            String attrtype = "";
            String[] ss = new String[1];
            ss[0] = attrname;
            int[] SDattrInfo = new int[2];
            if (hdf.SDattrinfo(sdid, i, ss, SDattrInfo))
            {
                attrname = ss[0];
                attrtype = hdf.HDgetNTdesc(SDattrInfo[0]);

                int l = MAX_NAME_LENGTH - attrname.length();
                if (l > 0) for(int j=0; j<l; j++) attrname += " ";
                else attrname = attrname.substring(0, MAX_NAME_LENGTH);
                textInfo += attrname;

                int attrSize = hdf.DFKNTsize(SDattrInfo[0])*SDattrInfo[1] + 1;
                byte[] buf = new byte[attrSize];
                if ( !hdf.SDreadattr(sdid, i, buf) ) continue;
                textInfo += doattrs( buf, attrSize, SDattrInfo[0], SDattrInfo[1]);
            }
        } // end of for (int i=0; i<sdsGlobalAttrNumber; i++)

        return textInfo;
    }

    /**
     *  display the HDF file SDS dataset attributes
     *
     *  @param sdid       the SDS identifier
     *  @param fileInfo   the SDS data information
     *  @param index      the index of the HDF data object
     *  @return           the string of the data attribute.
     *
     *  @exception ncsa.hdf.hdflib.HDFException 
     *             should be thrown for errors in the
     *             HDF library call, but is not yet implemented.
     *
     */
    private String getHdfSDSattr(int sdid, int[] fileInfo, int index)
        throws HDFException
    {
        String textInfo = "";

        int sdsid = hdf.SDselect(sdid,index);
        if (sdsid == HDFConstants.FAIL) return textInfo;

        // check if the given SDS is the dimension scale
        if (hdf.SDiscoordvar(sdsid)) {
            hdf.SDendaccess(sdsid);
            return textInfo;
        }

        // get datsset info.
        int [] SDInfo = new int[3];
        String datasetname = new String(" ");
        String[] ss = new String[1];
        ss[0] = datasetname;
        int  dimsize[] = new int[10];
        if (hdf.SDgetinfo(sdsid, ss, dimsize, SDInfo))
        {
            int datasetAttrNumber = SDInfo[2];
            int long_name_idx = hdf.SDfindattr(sdsid, "long_name");
            int [] SDattrInfo = new int[2];
            hdf.SDattrinfo(sdsid, long_name_idx, ss,SDattrInfo);

            textInfo = "\nDataset Attributes";
            textInfo += "\n\nName";
            for (int j=0; j<MAX_NAME_LENGTH-4; j++) textInfo += " ";
            //textInfo += " Type";
            //for (int j=0; j<MAX_NAME_LENGTH-5; j++) textInfo += " ";
            textInfo += " Value\n";
            for (int j=0; j<80; j++) textInfo += "_";

            // for each dataset attribute
            for (int kk=0; kk<datasetAttrNumber; kk++)
            {
                // get attribute info.
                textInfo += "\n";
                String attrname = "";
                String attrtype = "";
                ss[0] = attrname;
                if (hdf.SDattrinfo( sdsid, kk, ss, SDattrInfo))
                {
                    attrname = ss[0];
                    attrtype = hdf.HDgetNTdesc(SDattrInfo[0]);

                    int l = MAX_NAME_LENGTH - attrname.length();
                    if (l > 0) for(int j=0; j<l; j++) attrname += " ";
                    else attrname = attrname.substring(0, MAX_NAME_LENGTH);
                    textInfo += attrname;

                    int attrSize = hdf.DFKNTsize(SDattrInfo[0])*SDattrInfo[1] + 1;
                    byte[] buf = new byte[attrSize];
                    if ( !hdf.SDreadattr(sdsid, kk, buf) ) continue;
                    textInfo += doattrs( buf, attrSize, SDattrInfo[0], SDattrInfo[1]);
                }
            } // end of for (int kk=0; kk<datasetAttrNumber; kk++)
        }
        return textInfo;
    }

    /**
     *  define the data object
     *
     *  @param dataType  the type of the iamge data
     *  @param dataSize  the size of the image data array
     */
    private Object defineDataObject(int dataType, int dataSize)
    {
        Object data = null;

        switch(dataType)
        {
            // one byte data
            case HDFConstants.DFNT_CHAR:
            case HDFConstants.DFNT_UCHAR8:
            case HDFConstants.DFNT_UINT8:
            case HDFConstants.DFNT_INT8:
                data = new byte[dataSize];
                break;
            // short
            case HDFConstants.DFNT_INT16:
            case HDFConstants.DFNT_UINT16:
                data = new short[dataSize];
                break;
            case HDFConstants.DFNT_INT32:
            case HDFConstants.DFNT_UINT32:
                data = new int[dataSize];
                break;
            case HDFConstants.DFNT_INT64:
            case HDFConstants.DFNT_UINT64:
                data = new long[dataSize];
                break;
            case HDFConstants.DFNT_FLOAT32:
                data = new float[dataSize];
                break;
            case HDFConstants.DFNT_FLOAT64:
                data = new double[dataSize];
                break;
        }
        return data;
    }

    /** return the original image size */
    public Dimension getOriginalImageSize() { return originalImageSize; }

    /** return the requested image size */
    public Dimension getCurrentImageSize() { return selectedImageSize; }

    /** return the image arguments */
    public int[] getImageArgv() { return imageArgv; }

    /** return the iamge data range */
    public double[] getDataRange() { return dataRange; }

    /** return data for animation images */
    public byte[][] getImages() { return images; }

    /** return dimension units */
    public String[] getDimUnits() { return dimUnits; }

    /** return dimension types */
    public int[] getDimTypes() { return dimTypes; }

    /** return dimension sizes */
    public int[] getDimSizes() { return dimSizes; }

    /** return dimension scales */
    public Object[] getDimScales() { return dimScales; }

    /** return dimension names */
    public String[] getDimNames() { return dimNames; }

}
