package pupi.reader.restart;
import dataspork.tree.*;
import dataspork.reader.*;
import java.util.Stack;
import dataspork.interp.*;
import java.util.Enumeration;

public class RSDocHandler implements HDFDocHandler {
  protected DataTreeNode node;
  protected Stack stack;
  protected DataTreeNode root;
  protected DataTreeNode calc;
  protected DataTreeNode firstCalc; //for handling clones
  protected DataTreeNode grNode;
  protected DataTreeNode skNode;
  protected ScalarArraySource dataSource;
  protected DataTreeNode dataNode;
  protected ScalarArraySource wtDataSource;
  protected DataTreeNode wtNode;
  protected int dataSourceTreeLevel;
  protected DataTreeNode wtGroup;
  protected int scalarGroupTreeLevel;
  protected int treeLevel;
  protected int blocksize;
  protected boolean isClone;
  protected String filename;

  public RSDocHandler() {
    stack = new Stack();
    isClone = false;
  }

  public void setIsClone(boolean b) {isClone=b;}

  public DataTreeNode getDataTree() {
    return isClone ? firstCalc : root;
  }  

  public void startElement(HDFParser.Element e) {
    ++treeLevel;
    //Save the previous node on stack.
    if (node!=null) stack.push(node);
    //Deal with element by integer userTag.
    switch (e.userTag) {
    case UpiTags.UPI_LABEL:
    { 
      node.setName(e.readString());
      break;
    }
    case UpiTags.UPI_INDEX:
    { if (e.isVec()) {
        int[] temp = e.readIntegerArray();
        blocksize=temp[0];
      }
      else {
        node.setName(node+" "+e.readInteger());
      }
      break;
    }
    case UpiTags.UPI_SCALARS:
    { //Move scalars directly into calc node if present.
      if (calc!=null) {
        node.removeFromParent();
        node=calc;
        stack.pop();
        stack.push(node);
      } else {
        node.removeFromParent();
      }
      wtGroup = new DataTreeNode("weights");
      scalarGroupTreeLevel = treeLevel-1;
      node.add(wtGroup);
      break;
    }
    case UpiTags.UPI_SCALAR:
    { dataSource = new ScalarArraySource(e);
      dataSourceTreeLevel = treeLevel-1;
      node.setDataSource(dataSource);

      node.addInterpreter(new InterpreterClass(ScalarDataset.class,node));
      dataNode=node;
      wtDataSource = new ScalarArraySource(e);
      wtNode = new DataTreeNode("wt",wtDataSource); //Name set in endElement

      //InterpreterClass wtInterp = new InterpreterClass(ScalarDataset.class,wtNode);
      //wtNode.addInterpreter(wtInterp);
      //ScalarDataset wtInst = (ScalarDataset)wtInterp.getInstance();
      //wtInst.noReport();   // turn off the report function for weight scalars

      wtNode.addInterpreter(new InterpreterClass(ScalarDataset.class,wtNode));
      wtGroup.add(wtNode);
      break;
    }
    case UpiTags.UPI_BLOCKAVG: 
    { break;
    }
    case UpiTags.UPI_BLOCKSUM:
    {
      dataSource.addBlockSums(e,blocksize);
      break;
    }
    case UpiTags.UPI_BLOCKNORM:
    {
      wtDataSource.addBlockSums(e,blocksize);
      break;
    }
    case UpiTags.UPI_UNITS:
    { node.addAttribute(node.UNITS,e.readString());
      break;
    }
    case UpiTags.UPI_CALC:
    { //Move calculations to root node.
      node.removeFromParent();
      root.add(node);
      node.setName(e.toString());
      node.addAttribute("UpiTag",e.toString());
      dataSourceTreeLevel = -1;
      if (calc==null) firstCalc=node;
      calc=node;
      calc.addInterpreter(new InterpreterClass(ScalarGroup.class,node));
      break;
    } 
    case UpiTags.UPI_VM:
    case UpiTags.UPI_OBSERVABLES:
    case UpiTags.UPI_PATH:
    case UpiTags.UPI_RESTART:
    { //Remove from tree (continue parsing, then let garbage collection get it).
      node.removeFromParent();
      break;
    }
    case UpiTags.UPI_SETS:
    case UpiTags.UPI_BLOCKS:
    case UpiTags.UPI_TIME:
    case UpiTags.UPI_EPOCH:
    { //Do nothing.
      break;
    }
    case UpiTags.UPI_VECTORS:
    { //Make directories for pair distribution and structure factors.
      createNewNode(e);
      node.removeFromParent();
      skNode=new DataTreeNode("s(k)");
      grNode=new DataTreeNode("g(r)");
      root.add(skNode);
      root.add(grNode);
      break;
    }
    case UpiTags.UPI_SKPOINTS:
    { //Move skpoints to skNode.
      createNewNode(e);
      node.setName("k-grid");
      node.removeFromParent();
      node.setDataSource(new VectorSource(e));
      skNode.add(node);
      break;
    }
    case UpiTags.UPI_SK:
    { //Move structure factors to skNode.
      node.removeFromParent();
      skNode.add(node);
      node.addInterpreter(new InterpreterClass(StructureFactor.class,node));
      break;
    }
    case UpiTags.UPI_GR:
    { //Move pair distribution functions to grNode.
      node.removeFromParent();
      grNode.add(node);
      node.addInterpreter(new InterpreterClass(PairDistribution.class,node));
      break;
    }
    case UpiTags.UPI_DATA:
    {
      node.setDataSource(new VectorSource(e));
      //node.setDataSource(new ConstDataSource(
      //  new double[] {4,5,4,3,2,1},new int[] {6}, ConstDataSource.DOUBLE
      //));
      break;
    }
    case UpiTags.UPI_ERROR:
    { //For now, just ignore the error bars.
      break;
    }
    case UpiTags.UPI_NODE:
    case UpiTags.UPI_RNG:
    case UpiTags.UPI_GLOBAL_RNG:
    case UpiTags.UPI_EVIRIAL:
    { node.setName(e.toString());
      node.addAttribute("UpiTag",e.toString());
      break;
    }
    default: createNewNode(e);
    }
  }

  public void endElement(HDFParser.Element e) {
    --treeLevel;
    if (treeLevel<dataSourceTreeLevel && dataNode!=null) {
      wtNode.setName(dataNode.toString()+" weight"); //make wt names match data

      // set long name for easy window identification      
      dataNode.setLongName(dataNode.toString() + " ("+filename+")");
      wtNode.setLongName(dataNode.toString()+" weight"+" ("+filename+")");

      // turn off reporting for weight nodes
      if (wtNode.getNInterpreters() == 1) {
        Enumeration eInterp = wtNode.getInterpreters();
        InterpreterClass interp = (InterpreterClass)eInterp.nextElement();
        ScalarDataset in = (ScalarDataset)interp.getInstance();
        in.noReport();
      }

      // DEBUG
      //DataSource junkSource = dataNode.getDataSource();
      //if (junkSource == null) System.out.println("NULL");
      //double junkArray[] = (double[])junkSource.getData();
      //System.out.println("dataNode "+dataNode.toString()+" has length "+junkArray.length);
      //Enumeration eAttrib = dataNode.getAttributeLabels();
      //while (eAttrib.hasMoreElements()) {
      //  String aKey = (String)eAttrib.nextElement();
      //  System.out.println("attribute "+aKey+" = "+dataNode.getAttribute(aKey));
      //}
      // DEBUG


      dataNode.removeAllChildren(); //remove leftover groups from chunked data
      dataNode=null;
      wtNode=null;
      dataSource=null;
      wtDataSource=null;
    }
    if (treeLevel<scalarGroupTreeLevel && wtGroup!=null) {
      //move the wtGroup to the end of scalar group.
      ((DataTreeNode)wtGroup.getParent()).add(wtGroup);
      wtGroup=null;
    }
    if (treeLevel<0) calc=null;
    //Retrieve previous node from stack.
    node = stack.empty() ? null : (DataTreeNode)stack.pop();
  }

  protected void createNewNode(HDFParser.Element e) {
    DataTreeNode newNode = new DataTreeNode(e.toString());
    if (node!=null) {
      node.add(newNode);
    } else {
      root=newNode;
    }
    node = newNode;
  }

  public void startDocument() {
    treeLevel=0;
  }

  public void endDocument() {
    //Check if this was parallel data.
    root.addInterpreter(new InterpreterClass(ParallelGroup.class,root));
  }

  public void setFilename(String f) {
    filename = f;
  }

}
